#pragma once
#include <algorithm>
#include <string>
#include <unordered_map>
#include <vector>

#include "Deer/Components.h"
#include "Deer/Log.h"
#include "entt/entt.hpp"

#ifdef DEER_RENDER
#include "DeerRender/Render/FrameBuffer.h"
#include "DeerRender/SceneCamera.h"
#endif

namespace Deer {
	class Entity;
	using EntityMap = std::unordered_map<uint32_t, Entity>;

	class Environment {
	   public:
		Environment();
		~Environment();

		void clear();

#ifdef DEER_RENDER
		void render(SceneCamera& camera);
#endif

		Entity& getEntity(uint32_t id);
		Entity& createEntity(const std::string& name = std::string());
		Entity createEmptyEntity();

		// FEO
		uint32_t tryGetMainCamera();
		void setMainCamera(Entity& entity);

		Entity& getRoot();

	   public:
		entt::registry m_registry;
		EntityMap m_entities;

		uint32_t m_rootEntity = 0;
		uint32_t m_mainCamera = 0;

	   private:
		uint32_t m_idCreationOffset = 0;
		inline uint32_t pullEntityID() {
			m_idCreationOffset++;
			return m_idCreationOffset;
		}

		friend class Entity;
	};

	class Entity {
	   public:
		Entity() {}

		static Entity nullEntity;

		template <typename T, typename... Args>
		T& addComponent(Args&&... args) const {
			DEER_CORE_ASSERT(
			    !m_environment->m_registry.all_of<T>(m_entityHandle),
			    "Entity already have component {0}", typeid(T).name());

			return m_environment->m_registry.emplace<T>(
			    m_entityHandle, std::forward<Args>(args)...);
		}

		template <typename T>
		T& getComponent() const {
			DEER_CORE_ASSERT(
			    m_environment->m_registry.all_of<T>(m_entityHandle),
			    "Entity has no component {0}", typeid(T).name());

			return m_environment->m_registry.get<T>(m_entityHandle);
		}

		template <typename T>
		bool hasComponent() const {
			return m_environment->m_registry.all_of<T>(m_entityHandle);
		}

		template <typename T>
		void removeComponent() const {
			DEER_CORE_ASSERT(
			    m_environment->m_registry.all_of<T>(m_entityHandle),
			    "Entity does not have component {0}", typeid(T).name());

			m_environment->m_registry.remove<T>(m_entityHandle);
		}

		Entity& duplicate();
		void destroy();

		Entity& getParent();
		inline uint32_t getParentUID() { return m_parentUID; }

		// TODO, enable transfer entitys from difrent enviroments
		void setParent(Entity& parent);
		bool isDescendant(Entity& parent);

		uint32_t getParentUID() const { return m_parentUID; }
		uint32_t getUID() const { return m_entityUID; }

		Environment* getEnvironment() const { return m_environment; }
		std::vector<uint32_t>& getChildren();

		bool isRoot() { return m_isRoot; }
		glm::mat4 getWorldMatrix();
		glm::mat4 getRelativeMatrix();

		void updateInternalVars();

		inline bool isValid() const {
			return m_entityUID != 0 && m_environment != nullptr &&
			       m_environment->m_registry.valid(m_entityHandle);
		}
		inline bool operator==(const Entity& b) const {
			return m_environment == b.m_environment &&
			       m_entityUID == b.m_entityUID;
		}

	   private:
		Entity(entt::entity handle, Environment* scene);
		bool removeChild(Entity& child);

		entt::entity m_entityHandle = entt::null;
		Environment* m_environment = nullptr;
		uint32_t m_entityUID = 0;
		uint32_t m_parentUID = 0;
		bool m_isRoot = false;

		friend class Environment;
		friend class std::unordered_map<uint32_t, Entity>;
	};
}  // namespace Deer
